@@ -1,10 +1,11 @@
 __author__ = "Johannes Köster"
-__copyright__ = "Copyright 2021, Johannes Köster"
+__copyright__ = "Copyright 2022, Johannes Köster"
 __email__ = "johannes.koester@uni-due.de"
 __license__ = "MIT"
 
 import os
 import re
+import types
 from snakemake.path_modifier import PATH_MODIFIER_FLAG
 import sys
 import inspect
@@ -55,7 +56,14 @@ from snakemake.exceptions import (
     IncompleteCheckpointException,
 )
 from snakemake.logging import logger
-from snakemake.common import Mode, ON_WINDOWS, lazy_property, TBDString
+from snakemake.common import (
+    Mode,
+    ON_WINDOWS,
+    get_function_params,
+    get_input_function_aux_params,
+    lazy_property,
+    TBDString,
+)
 import snakemake.io
 
 
@@ -103,6 +111,7 @@ class Rule:
             self.script = None
             self.notebook = None
             self.wrapper = None
+            self.template_engine = None
             self.cwl = None
             self.norun = False
             self.is_handover = False
@@ -110,8 +119,12 @@ class Rule:
             self.is_checkpoint = False
             self.restart_times = 0
             self.basedir = None
-            self.path_modifer = None
+            self.input_modifier = None
+            self.output_modifier = None
+            self.log_modifier = None
+            self.benchmark_modifier = None
             self.ruleinfo = None
+            self.module_globals = None
         elif len(args) == 1:
             other = args[0]
             self.name = other.name
@@ -152,6 +165,7 @@ class Rule:
             self.script = other.script
             self.notebook = other.notebook
             self.wrapper = other.wrapper
+            self.template_engine = other.template_engine
             self.cwl = other.cwl
             self.norun = other.norun
             self.is_handover = other.is_handover
@@ -159,8 +173,12 @@ class Rule:
             self.is_checkpoint = other.is_checkpoint
             self.restart_times = other.restart_times
             self.basedir = other.basedir
-            self.path_modifier = other.path_modifier
+            self.input_modifier = other.input_modifier
+            self.output_modifier = other.output_modifier
+            self.log_modifier = other.log_modifier
+            self.benchmark_modifier = other.benchmark_modifier
             self.ruleinfo = other.ruleinfo
+            self.module_globals = other.module_globals
 
     def dynamic_branch(self, wildcards, input=True):
         def get_io(rule):
@@ -230,9 +248,14 @@ class Rule:
                 if len(set(values)) == 1
             )
             # TODO have a look into how to concretize dependencies here
-            branch._input, _, branch.dependencies = branch.expand_input(
+            branch._input, _, branch.dependencies, incomplete = branch.expand_input(
                 non_dynamic_wildcards
             )
+            assert not incomplete, (
+                "bug: dynamic branching resulted in incomplete input files, "
+                "please file an issue on https://github.com/snakemake/snakemake"
+            )
+
             branch._output, _ = branch.expand_output(non_dynamic_wildcards)
 
             resources = branch.expand_resources(non_dynamic_wildcards, branch._input, 1)
@@ -267,6 +290,10 @@ class Rule:
     def is_wrapper(self):
         return self.wrapper is not None
 
+    @property
+    def is_template_engine(self):
+        return self.template_engine is not None
+
     @property
     def is_cwl(self):
         return self.cwl is not None
@@ -330,7 +357,9 @@ class Rule:
         if isinstance(benchmark, Path):
             benchmark = str(benchmark)
         if not callable(benchmark):
-            benchmark = self.apply_path_modifier(benchmark, property="benchmark")
+            benchmark = self.apply_path_modifier(
+                benchmark, self.benchmark_modifier, property="benchmark"
+            )
             benchmark = self._update_item_wildcard_constraints(benchmark)
 
         self._benchmark = IOFile(benchmark, rule=self)
@@ -453,9 +482,9 @@ class Rule:
                 )
             seen[value] = name or idx
 
-    def apply_path_modifier(self, item, property=None):
-        assert self.path_modifier is not None
-        apply = partial(self.path_modifier.modify, property=property)
+    def apply_path_modifier(self, item, path_modifier, property=None):
+        assert path_modifier is not None
+        apply = partial(path_modifier.modify, property=property)
 
         assert not callable(item)
         if isinstance(item, dict):
@@ -510,9 +539,14 @@ class Rule:
             if isinstance(item, _IOFile) and item.rule and item in item.rule.output:
                 rule_dependency = item.rule
 
-            item = self.apply_path_modifier(
-                item, property="output" if output else "input"
-            )
+            if output:
+                path_modifier = self.output_modifier
+                property = "output"
+            else:
+                path_modifier = self.input_modifier
+                property = "input"
+
+            item = self.apply_path_modifier(item, path_modifier, property=property)
 
             # Check to see that all flags are valid
             # Note that "remote", "dynamic", and "expand" are valid for both inputs and outputs.
@@ -525,6 +559,8 @@ class Rule:
                         "directory",
                         "touch",
                         "pipe",
+                        "service",
+                        "ensure",
                     ]:
                         logger.warning(
                             "The flag '{}' used in rule {} is only valid for outputs, not inputs.".format(
@@ -582,6 +618,7 @@ class Rule:
                         self.workflow.current_basedir.join(report_obj.caption),
                         report_obj.category,
                         report_obj.subcategory,
+                        report_obj.labels,
                         report_obj.patterns,
                         report_obj.htmlindex,
                     )
@@ -665,7 +702,7 @@ class Rule:
             item = str(item)
         if isinstance(item, str) or callable(item):
             if not callable(item):
-                item = self.apply_path_modifier(item, property="log")
+                item = self.apply_path_modifier(item, self.log_modifier, property="log")
                 item = self._update_item_wildcard_constraints(item)
 
             self.log.append(IOFile(item, rule=self) if isinstance(item, str) else item)
@@ -699,17 +736,32 @@ class Rule:
         wildcards,
         incomplete_checkpoint_func=lambda e: None,
         raw_exceptions=False,
-        **aux_params
+        groupid=None,
+        **aux_params,
     ):
         incomplete = False
         if isinstance(func, _IOFile):
             func = func._file.callable
         elif isinstance(func, AnnotatedString):
             func = func.callable
-        sig = inspect.signature(func)
-        _aux_params = {k: v for k, v in aux_params.items() if k in sig.parameters}
+
+        if "groupid" in get_function_params(func):
+            if groupid is not None:
+                aux_params["groupid"] = groupid
+            else:
+                # Return empty list of files and incomplete marker
+                # the job will be reevaluated once groupids have been determined
+                return [], True
+
+        _aux_params = get_input_function_aux_params(func, aux_params)
+
         try:
             value = func(Wildcards(fromdict=wildcards), **_aux_params)
+            if isinstance(value, types.GeneratorType):
+                # generators should be immediately collected here,
+                # otherwise we would miss any exceptions and
+                # would have to capture them again later.
+                value = list(value)
         except IncompleteCheckpointException as e:
             value = incomplete_checkpoint_func(e)
             incomplete = True
@@ -739,11 +791,13 @@ class Rule:
         mapping=None,
         no_flattening=False,
         aux_params=None,
-        apply_path_modifier=True,
+        path_modifier=None,
         property=None,
         incomplete_checkpoint_func=lambda e: None,
         allow_unpack=True,
+        groupid=None,
     ):
+        incomplete = False
         if aux_params is None:
             aux_params = dict()
         for name, item in olditems._allitems():
@@ -759,37 +813,39 @@ class Rule:
                     wildcards,
                     incomplete_checkpoint_func=incomplete_checkpoint_func,
                     is_unpack=is_unpack,
-                    **aux_params
+                    groupid=groupid,
+                    **aux_params,
                 )
-                if apply_path_modifier:
-                    item = self.apply_path_modifier(item, property=property)
 
             if is_unpack and not incomplete:
                 if not allow_unpack:
                     raise WorkflowError(
                         "unpack() is not allowed with params. "
                         "Simply return a dictionary which can be directly ."
-                        "used, e.g. via {params[mykey]}."
+                        "used, e.g. via {params[mykey]}.",
+                        rule=self,
                     )
                 # Sanity checks before interpreting unpack()
                 if not isinstance(item, (list, dict)):
                     raise WorkflowError(
-                        "Can only use unpack() on list and dict", rule=self
+                        f"Can only use unpack() on list and dict, but {item} was returned.",
+                        rule=self,
                     )
                 if name:
                     raise WorkflowError(
-                        "Cannot combine named input file with unpack()", rule=self
+                        f"Cannot combine named input file (name {name}) with unpack()",
+                        rule=self,
                     )
                 # Allow streamlined code with/without unpack
                 if isinstance(item, list):
-                    pairs = zip([None] * len(item), item)
+                    pairs = zip([None] * len(item), item, [_is_callable] * len(item))
                 else:
                     assert isinstance(item, dict)
-                    pairs = item.items()
+                    pairs = [(name, item, _is_callable) for name, item in item.items()]
             else:
-                pairs = [(name, item)]
+                pairs = [(name, item, _is_callable)]
 
-            for name, item in pairs:
+            for name, item, from_callable in pairs:
                 is_iterable = True
                 if not_iterable(item) or no_flattening:
                     item = [item]
@@ -801,8 +857,14 @@ class Rule:
                         and not isinstance(item_, Path)
                     ):
                         raise WorkflowError(
-                            "Function did not return str or list " "of str.", rule=self
+                            "Function did not return str or list of str.", rule=self
                         )
+
+                    if from_callable and path_modifier is not None and not incomplete:
+                        item_ = self.apply_path_modifier(
+                            item_, path_modifier, property=property
+                        )
+
                     concrete = concretize(item_, wildcards, _is_callable)
                     newitems.append(concrete)
                     if mapping is not None:
@@ -813,8 +875,9 @@ class Rule:
                         name, start, end=len(newitems) if is_iterable else None
                     )
                     start = len(newitems)
+        return incomplete
 
-    def expand_input(self, wildcards):
+    def expand_input(self, wildcards, groupid=None):
         def concretize_iofile(f, wildcards, is_from_callable):
             if is_from_callable:
                 if isinstance(f, Path):
@@ -839,14 +902,16 @@ class Rule:
         input = InputFiles()
         mapping = dict()
         try:
-            self._apply_wildcards(
+            incomplete = self._apply_wildcards(
                 input,
                 self.input,
                 wildcards,
                 concretize=concretize_iofile,
                 mapping=mapping,
                 incomplete_checkpoint_func=handle_incomplete_checkpoint,
+                path_modifier=self.input_modifier,
                 property="input",
+                groupid=groupid,
             )
         except WildcardError as e:
             raise WildcardError(
@@ -869,7 +934,7 @@ class Rule:
         for f in input:
             f.check()
 
-        return input, mapping, dependencies
+        return input, mapping, dependencies, incomplete
 
     def expand_params(self, wildcards, input, output, resources, omit_callable=False):
         def concretize_param(p, wildcards, is_from_callable):
@@ -886,7 +951,6 @@ class Rule:
         def handle_incomplete_checkpoint(exception):
             """If checkpoint is incomplete, target it such that it is completed
             before this rule gets executed."""
-            print(exception.targetfile)
             if exception.targetfile in input:
                 return TBDString()
             else:
@@ -908,7 +972,6 @@ class Rule:
                 omit_callable=omit_callable,
                 allow_unpack=False,
                 no_flattening=True,
-                apply_path_modifier=False,
                 property="params",
                 aux_params={
                     "input": input._plainstrings(),
@@ -957,7 +1020,12 @@ class Rule:
 
         try:
             self._apply_wildcards(
-                log, self.log, wildcards, concretize=concretize_logfile, property="log"
+                log,
+                self.log,
+                wildcards,
+                concretize=concretize_logfile,
+                path_modifier=self.log_modifier,
+                property="log",
             )
         except WildcardError as e:
             raise WildcardError(
@@ -1005,7 +1073,7 @@ class Rule:
                         attempt=attempt,
                         incomplete_checkpoint_func=lambda e: 0,
                         raw_exceptions=True,
-                        **aux
+                        **aux,
                     )
                 except (Exception, BaseException) as e:
                     raise InputFunctionException(e, rule=self, wildcards=wildcards)
@@ -1016,13 +1084,23 @@ class Rule:
 
             if not isinstance(res, int) and not isinstance(res, str):
                 raise WorkflowError(
-                    "Resources function did not return int, float (floats are "
-                    "rouded to the nearest integer), or str.",
+                    f"Resource {name} is neither int, float(would be rounded to nearest int), or str.",
                     rule=self,
                 )
-            if isinstance(res, int):
-                global_res = self.workflow.global_resources.get(name, res)
-                if global_res is not None:
+
+            global_res = self.workflow.global_resources.get(name)
+            if global_res is not None:
+                if not isinstance(res, TBDString) and type(res) != type(global_res):
+                    global_type = (
+                        "an int" if isinstance(global_res, int) else type(global_res)
+                    )
+                    raise WorkflowError(
+                        f"Resource {name} is of type {type(res).__name__} but global resource constraint "
+                        f"defines {global_type} with value {global_res}. "
+                        "Resources with the same name need to have the same types (int, float, or str are allowed).",
+                        rule=self,
+                    )
+                if isinstance(res, int):
                     res = min(global_res, res)
             return res
 
@@ -1192,7 +1270,23 @@ class RuleProxy:
 
     @lazy_property
     def input(self):
-        return self.rule.input._stripped_constraints()
+        def modify_callable(item):
+            if is_callable(item):
+                # For callables ensure that the rule's original path modifier is applied as well.
+
+                def inner(wildcards):
+                    return self.rule.apply_path_modifier(
+                        item(wildcards), self.rule.input_modifier, property="input"
+                    )
+
+                return inner
+            else:
+                # For strings, the path modifier has been already applied.
+                return item
+
+        return InputFiles(
+            toclone=self.rule.input, strip_constraints=True, custom_map=modify_callable
+        )
 
     @lazy_property
     def params(self):
